# If threeml is not installed, we create our own log,
# otherwise, just print to the 3ML one!


import logging
import logging.handlers as handlers
from pathlib import Path

from astromodels.utils.configuration import astromodels_config
from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

from .file_utils import _get_data_file_path

try:

    from threeML.config.config import threeML_config

    has_threeml = True

except:

    has_threeml = False


def get_path_of_log_dir():

    # we use the 3ML log path to simplify things
    # a more clever solution could be found

    if has_threeml:

        user_log: Path = Path(threeML_config.logging.path).expanduser()

    else:

        user_log: Path = Path(astromodels_config.logging.path).expanduser()

    # Create it if doesn't exist
    if not user_log.exists():

        user_log.mkdir(parents=True)

    return user_log


_log_file_names = ["usr.log", "dev.log"]


def get_path_of_log_file(log_file: str) -> Path:
    """
    returns the path of the log files
    """
    assert log_file in _log_file_names, f"{log_file} is not one of {_log_file_names}"

    return get_path_of_log_dir() / log_file


class LogFilter(object):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno != self.__level


# now create the developer handler that rotates every day and keeps
# 10 days worth of backup
astromodels_dev_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("dev.log"), when="D", interval=1, backupCount=10
)

# lots of info written out

_dev_formatter = logging.Formatter(
    "%(asctime)s | %(name)s | %(levelname)s| %(funcName)s | %(lineno)d | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

astromodels_dev_log_handler.setFormatter(_dev_formatter)
astromodels_dev_log_handler.setLevel(logging.DEBUG)
# now set up the usr log which will save the info

astromodels_usr_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("usr.log"), when="D", interval=1, backupCount=10
)

astromodels_usr_log_handler.setLevel(logging.INFO)

# lots of info written out
_usr_formatter = logging.Formatter(
    "%(asctime)s | %(levelname)s | %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)

astromodels_usr_log_handler.setFormatter(_usr_formatter)


_console_formatter = logging.Formatter(
    " %(message)s",
    datefmt="%H:%M:%S",
)

mytheme = Theme().read(_get_data_file_path("log_theme.ini"))
console = Console(theme=mytheme)


astromodels_console_log_handler = RichHandler(
    level="INFO", rich_tracebacks=True, markup=True, console=console
)
astromodels_console_log_handler.setFormatter(_console_formatter)
astromodels_console_log_handler.setLevel("INFO")

warning_filter = LogFilter(logging.WARNING)


def silence_warnings():
    """
    supress warning messages in console and file usr logs
    """

    astromodels_usr_log_handler.addFilter(warning_filter)
    astromodels_console_log_handler.addFilter(warning_filter)


def activate_warnings():
    """
    supress warning messages in console and file usr logs
    """

    astromodels_usr_log_handler.removeFilter(warning_filter)
    astromodels_console_log_handler.removeFilter(warning_filter)


def update_logging_level(level):

    astromodels_console_log_handler.setLevel(level)


def setup_logger(name):

    # A logger with name name will be created
    # and then add it to the print stream
    log = logging.getLogger(name)

    # this must be set to allow debug messages through
    log.setLevel(logging.DEBUG)

    # add the handlers

    log.addHandler(astromodels_dev_log_handler)

    log.addHandler(astromodels_console_log_handler)

    log.addHandler(astromodels_usr_log_handler)

    # we do not want to duplicate teh messages in the parents
    log.propagate = False

    return log



import collections
import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from astromodels.utils.logging import setup_logger

from .tree import Node

log = setup_logger(__name__)

# Exception for when a parameter is out of its bounds
class SettingUnknownValue(RuntimeError):
    pass


class PropertyBase(Node):

    def __init__(self,
                 name: str,
                 desc: str,
                 value: Optional[str] = None,
                 allowed_values: Optional[List[str]] = None,
                 defer: bool = False,
                 eval_func: Optional[str] = None
                 ):

        # Make this a node

        Node.__init__(self, name)

        self._allowed_values: Optional[List[str]] = allowed_values
        self._defer: bool = defer
        self._eval_func: Optional[str] = eval_func
        
        if (value is None) and (not self._defer):

            log.error(f"property {name} was given no initial value but is NOT deferred")

        # now we set the value

        self.value = value

        self.__doc__ = desc

        self._desc = desc

    def _get_value(self) -> Any:
        """
        Return current parameter value
        """

        log.debug(f"accessing the property {self.name} with value {self._internal_value}")
        
        return self._internal_value
    

    def _set_value(self, new_value) -> None:
        """
        Sets the current value of the parameter, ensuring that it is within the allowed range.
        """

        if (self._defer) and (new_value is None):

                # this is ok
                pass

        elif self._allowed_values is not None:

            if not new_value in self._allowed_values:

                log.error(f"{self.name} can only take the values {','.join(self._allowed_values)} not {new_value}")

                raise SettingUnknownValue()
                

        self._internal_value = new_value
                                       
        # if there is an eval func value
        # then we need to execute the function
        # on the parent

        if (self._internal_value == "_tmp") and self._defer:

            # do not execute in this mode
            
            return 

        if self._eval_func is not None:

            # if there is a parent
            if self._parent is not None:

                if self._parent.name == "composite":
                    # ok, we have a composite function

                    func_idx = int(self._name.split("_")[-1]) - 1

                    getattr(self._parent._functions[func_idx], str(self._eval_func))()

                else:
                
                    getattr(self._parent, str(self._eval_func))()

            # other wise this will run when the parent is set
                
        
        

    value = property(
        _get_value,
        _set_value,
        doc=
        "Get and sets the current value for the propert",)

    def _set_parent(self, parent):

        # we intecept here becuase we want
        # to make sure the eval works
        
        super(PropertyBase, self)._set_parent(parent)

        # now we want to update because we have a parent
        self.value = self._internal_value
        

    
    @property
    def is_deferred(self) -> bool:
        return self._defer

    @property
    def description(self) -> Optional[str]:
        """
        Return a description of this parameter

        :return: a string cointaining a description of the meaning of this parameter
        """
        return self._desc
    
    def duplicate(self) -> "FunctionProperty":
        """
        Returns an exact copy of the current property
        """

        # Deep copy everything to make sure that there are no ties between the new instance and the old one

        new_property = copy.deepcopy(self)

        return new_property

    def _repr__base(self, rich_output):  # pragma: no cover

        raise NotImplementedError(
            "You need to implement this for the actual Property class")


    @staticmethod
    def _to_python_type(variable):
        """
        Returns the value in the variable handling also np.array of one element

        :param variable: input variable
        :return: the value of the variable having a python type (int, float, ...)
        """

        # Assume variable is a np.array, fall back to the case where variable is already a primitive type

        try:

            return variable.item()

        except AttributeError:

            return variable

    
    def to_dict(self, minimal=False) -> Dict[str, Any]:
        """Returns the representation for serialization"""

        data = collections.OrderedDict()

        if minimal:

            # In the minimal representation we just output the value

            data["value"] = self._to_python_type(self.value)

        else:

            # In the complete representation we output everything is needed to re-build the object

            data["value"] = str(self.value)
            data["desc"] = str(self._desc)
            data["allowed values"] = self._to_python_type(self._allowed_values)
            data["defer"] = self._to_python_type(self._defer)
            data["function"] = str(self._eval_func)

        return data

    
        
class FunctionProperty(PropertyBase):

        def __init__(self,
                     name: str,
                     desc: str,
                     value: Optional[str] = None,
                     allowed_values: Optional[List[Any]] = None,
                     defer: bool = False,
                     eval_func: Optional[str] = None
                     
                     ):

            super(FunctionProperty, self).__init__(name=name,desc=desc,
                                                   value=value,
                                                   allowed_values=allowed_values,
                                                   defer=defer,
                                                   eval_func=eval_func)

            
        def _repr__base(self, rich_output=False):

            representation = (
                f"Property {self.name} = {self.value}\n"
                f"(allowed values = {'all' if self._allowed_values is None else ' ,'.join(self._allowed_values)})")
                
            return representation
            



import collections
import itertools
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Type

from astromodels.utils.logging import setup_logger

from .cpickle_compatibility_layer import cPickle

log = setup_logger(__name__)


# This is necessary for pickle to be able to reconstruct a NewNode class (or derivate)
# during unpickling
class NewNodeUnpickler(object):
    def __call__(self, cls):

        instance = cls.__new__(cls)

        return instance


@dataclass(repr=False, unsafe_hash=True)
class NodeBase:
    _name: str
    _parent: Optional[Type["NodeBase"]] = field(repr=False, default=None)
    _children: Dict[str, Type["NodeBase"]] = field(
        default_factory=dict, repr=False, compare=False
    )
    _path: Optional[str] = field(repr=False, default="")

    # The next 3 methods are *really* necessary for anything to work

    def __reduce__(self):

        state = {}
        state["parent"] = self._get_parent()
        state["path"] = self._path
        state["children"] = self._get_children()
        state["child_names"] = [child.name for child in state["children"]]
        state["name"] = self._name
        state["__dict__"] = self.__dict__

        return NewNodeUnpickler(), (self.__class__,), state

    def __setstate__(self, state) -> None:

        self._children = {}

        # Set the name of this node

        self._name = state["name"]

        # Set the parent

        self._parent = state["parent"]

        # set the path

        self._path = state["path"]

        # Set the children
        # do this manually to avoid recursion
        # issues for children not yet built

        for child, name in zip(state["children"], state["child_names"]):

            self._children[name] = child

        # Restore everything else

        for k in state["__dict__"]:

            self.__dict__[k] = state["__dict__"][k]

    # This is necessary for copy.deepcopy to work
    def __deepcopy__(self, memodict={}):

        return cPickle.loads(cPickle.dumps(self))

    def _add_child(self, child: Type["NodeBase"]) -> None:

        if not isinstance(child, NodeBase):

            log.error(f"{child} is not of type Node")

            raise TypeError()

        log.debug(f"adding child {child._name}")

        if child._name not in self._children:

            # add the child to the dict

            self._children[child._name] = child

            # set the parent

            child._set_parent(self)

            # now go through and make sure
            # all the children know about the
            # new parent recursively

            child._update_child_path()

        else:

            log.error(f"A child with name {child._name} already exists")

            raise AttributeError()

    def _add_children(self, children: List[Type["NodeBase"]]) -> None:

        for child in children:
            self._add_child(child)

    def _remove_child(self, name: str, delete: bool = True) -> Optional["NodeBase"]:
        """
        return a child
        """

        # this kills the child

        if delete:

            del self._children[name]

            # return none

        # we want to keep the child
        # but orphan it

        else:

            child = self._children.pop(name)

            # now orphan the child

            child._orphan()

            # we want to get the
            # orphan back

            return child

            # return self._children.pop(name)

    def _orphan(self) -> None:
        """
        This will disconnect the current node from its parent
        and inform all the children about the change
        """

        # disconnect from parent

        self._parent = None

        # be nice to the kids and tell them

        self._update_child_path()

    def _set_parent(self, parent: Type["NodeBase"]) -> None:
        """
        set the parent and update path
        """

        self._parent = parent

        parent_path = self._parent._get_path()

        if parent_path == "__root__":

            self._path = f"{self._name}"

        else:

            self._path = f"{parent_path}.{self._name}"

        log.debug(f"path is now: {self._path}")

    def _get_child(self, name: str) -> "NodeBase":
        """
        return a child object
        """

        return self._children[name]

    def _has_child(self, name: str) -> bool:
        """
        is this child (name) in the tree
        """
        return name in self._children

    def _get_children(self) -> Tuple["NodeBase"]:
        """
        return a tuple of children
        """

        return tuple(self._children.values())

    def _get_child_from_path(self, path: str) -> "NodeBase":
        """
        get a child from a string path
        """
        nodes = path.split(".")
        _next = self
        for node in nodes:
            _next = _next._get_child(node)

        return _next

    def __getitem__(self, key) -> "NodeBase":
        return self._get_child_from_path(key)

    def _recursively_gather_node_type(self, node, node_type) -> Dict[str, "NodeBase"]:

        instances = collections.OrderedDict()

        for child in node._get_children():

            # log.debug(f"on child {child._name}")

            if isinstance(child, node_type):

                path = child._get_path()

                # log.debug(f"on child {path}")

                instances[path] = child

                for sub_child in child._get_children():

                    instances.update(
                        self._recursively_gather_node_type(sub_child, node_type)
                    )

            else:

                instances.update(self._recursively_gather_node_type(child, node_type))

        return instances

    def _get_parent(self) -> "NodeBase":
        return self._parent

    def _get_path(self) -> "str":
        """
        returns the str path of this node
        """
        if self._parent is not None:
            return self._path

        else:
            return self._name

    def _root(self, source_only: bool = False) -> "NodeBase":
        """
        returns the root of the node, will stop at the source
        if source_only is set to True
        """
        if self.is_root:
            return self

        else:

            current_node = self

            # recursively walk up the tree to
            # the root

            while True:

                parent = current_node._parent

                if source_only:

                    if parent.name == "__root__":

                        return current_node

                current_node = current_node._parent

                if current_node.is_root:

                    return current_node

    @property
    def path(self) -> str:
        return self._get_path()

    def _update_child_path(self) -> None:
        """

        Update the path of all children recursively.
        This is needed if the name is changed

        :returns:

        """
        # recursively update the path

        for name, child in self._children.items():

            child._path = f"{child._parent._get_path()}.{child._name}"

            if not child.is_leaf:

                child._update_child_path()

    def _change_name(self, name: str, clear_parent: bool = False) -> None:
        """
        change the name of this node. This will have to update the
        children about the change. if clear_parent is provided, then
        the parent is removed
        """
        self._name = name

        if (self._parent is not None) and (not clear_parent):

            #            self._parent._children.pop(self._name)
            self._set_parent(self._parent)

        # update all the children
        self._update_child_path()

    @property
    def is_root(self) -> bool:
        """
        is this the root of the tree
        """
        return self._parent is None

    @property
    def is_leaf(self) -> bool:
        """
        is this a a leaf of the tree
        """
        if len(self._children) == 0:
            return True

        else:

            return False

    @property
    def name(self) -> str:
        return self._name

    def __getattr__(self, name):

        if name in self._children:

            return self._children[name]

        else:

            # log.error(f"Accessing an element {name} of the node that does not exist")

            raise AttributeError(
                f"Accessing an element {name} of the node that does not exist"
            )

            # return super(NodeBase).__getattr__(name)

    def __setattr__(self, name, value):

        ### We cannot change a node
        ### but if the node has a value
        ### attribute, we want to call that

        if "_children" in self.__dict__:
            if name in self._children:

                if "_internal_value" in self._children[name].__dict__:

                    if not self._children[name].is_leaf:

                        log.warning(
                            f"Trying to set the value of a linked parameter ({name}) directly has no effect "
                        )

                        return

                    else:
                        # ok, this is likely  parameter

                        self._children[name].value = value

                else:

                    # this is going to be a node which
                    # we are not allowed to erase

                    # log.error(f"Accessing an element {name} of the node that does not exist")

                    raise AttributeError(
                        f"Accessing an element {name} of the node that does not exist"
                    )
            else:
                return super().__setattr__(name, value)
        else:

            return super().__setattr__(name, value)

    def plot_tree(self) -> None:
        """
        this plots the tree to the
        screen
        """

        print(self._plot_tree_console())

    def _plot_tree_console(self):

        if self.is_leaf:
            return self.name

        child_strs = [
            child._plot_tree_console() for child in list(self._children.values())
        ]
        child_widths = [block_width(s) for s in child_strs]

        # How wide is this block?
        display_width = max(len(self.name), sum(child_widths) + len(child_widths) - 1)

        # Determines midpoints of child blocks
        child_midpoints = []
        child_end = 0
        for width in child_widths:
            child_midpoints.append(child_end + (width // 2))
            child_end += width + 1

        # Builds up the brace, using the child midpoints
        brace_builder = []
        for i in range(display_width):
            if i < child_midpoints[0] or i > child_midpoints[-1]:
                brace_builder.append(" ")
            elif i in child_midpoints:
                brace_builder.append("+")
            else:
                brace_builder.append("-")
        brace = "".join(brace_builder)

        name_str = "{:^{}}".format(self.name, display_width)
        below = stack_str_blocks(child_strs)

        return name_str + "\n" + brace + "\n" + below


def block_width(block) -> int:
    try:
        return block.index("\n")
    except ValueError:
        return len(block)


def stack_str_blocks(blocks) -> str:
    """Takes a list of multiline strings, and stacks them horizontally.

    For example, given 'aaa\naaa' and 'bbbb\nbbbb', it returns
    'aaa bbbb\naaa bbbb'.  As in:

    'aaa  +  'bbbb   =  'aaa bbbb
     aaa'     bbbb'      aaa bbbb'

    Each block must be rectangular (all lines are the same length), but blocks
    can be different sizes.
    """
    builder = []
    block_lens = [block_width(bl) for bl in blocks]
    split_blocks = [bl.split("\n") for bl in blocks]

    for line_list in itertools.zip_longest(*split_blocks, fillvalue=None):
        for i, line in enumerate(line_list):
            if line is None:
                builder.append(" " * block_lens[i])
            else:
                builder.append(line)
            if i != len(line_list) - 1:
                builder.append(" ")  # Padding
        builder.append("\n")

    return "".join(builder[:-1])



import ctypes
import math

import numba as nb
import numpy as np


@nb.vectorize
def _expm1(x):
    
    return math.expm1(x)

@nb.vectorize
def _exp(x):
    
    return math.exp(x)


@nb.vectorize
def _sqrt(x):
    
    return math.sqrt(x)


@nb.vectorize
def _pow(x, y):

    return math.pow(x, y)

_cache_functions = False

@nb.vectorize
def _log(x):

    return math.log(x)

@nb.vectorize
def _log10(x):

    return math.log10(x)



@nb.njit(fastmath=True, cache=_cache_functions)
def plaw_eval(x, K, index, piv):

    out = np.power(x / piv, index)

    return K * out


@nb.njit(fastmath=True, cache=_cache_functions)
def plaw_flux_norm(index, a, b):
    """
    energy flux power law
    """
    # use the limit of the
    if index != -2.0:

        dp2 = 2 + index

        intflux = (math.pow(b, dp2) - math.pow(a, dp2)) / dp2
    else:

        intflux = - math.log(a/b)

    return intflux


@nb.njit(fastmath=True, cache=_cache_functions)
def cplaw_eval(x, K, xc, index, piv):

    n = x.shape[0]
    out = np.empty(n)

    for i in range(n):
        # Compute it in logarithm to avoid roundoff errors, then raise it
        log_v = index * np.log(x[i] / piv) - (x[i] / xc)
        out[i] = K * np.exp(log_v)

    return out


@nb.njit(fastmath=True, cache=_cache_functions)
def cplaw_inverse_eval(x, K, b, index, piv):

    n = x.shape[0]
    out = np.empty(n)

    for i in range(n):
        # Compute it in logarithm to avoid roundoff errors, then raise it
        log_v = index * np.log(x[i] / piv) - x[i] * b
        out[i] = K * np.exp(log_v)

    return out


@nb.njit(fastmath=True, cache=_cache_functions)
def super_cplaw_eval(x, K, piv, index, xc, gamma):

    n = x.shape[0]
    out = np.empty(n)

    for i in range(n):

        log_v = index * np.log(x[i] / piv) - gamma*(x[i] / xc)

        out[i] = K * np.exp(log_v)

    return out


@nb.njit(fastmath=True, cache=_cache_functions)
def band_eval(x, K, alpha, beta, E0, piv):

    n = x.shape[0]
    out = np.empty(n)

    break_point = (alpha - beta) * E0

    factor_ab = np.exp(beta - alpha) * \
        math.pow(break_point / piv, alpha - beta)

    for idx in range(n):

        if x[idx] < break_point:
            out[idx] = K * math.pow(x[idx] / piv, alpha) * np.exp(-x[idx] / E0)

        else:

            out[idx] = K * factor_ab * math.pow(x[idx] / piv, beta)

    return out


@nb.njit(fastmath=True, cache=_cache_functions)
def bplaw_eval(x, K, xb, alpha, beta, piv):

    n = x.shape[0]
    out = np.empty(n)

    factor = math.pow(xb / piv, alpha - beta)

    for idx in range(n):

        if x[idx] < xb:

            out[idx] = K * math.pow(x[idx] / piv, alpha)
        else:

            out[idx] = K * factor * math.pow(x[idx] / piv, beta)

    return out


@nb.njit(fastmath=True, cache=_cache_functions)
def sbplaw_eval(x, K, alpha, be, bs, beta, piv):

    n = x.shape[0]
    out = np.zeros(n)

    B = 0.5 * (alpha + beta)
    M = 0.5 * (beta - alpha)

    arg_piv = math.log10(piv / be) / bs

    log2 = math.log(2.0)

    Mbs = M * bs

    if arg_piv < -6.0:
        pcosh_piv = Mbs * (-arg_piv - log2)

    elif arg_piv > 4.0:

        pcosh_piv = Mbs * (arg_piv - log2)
    else:

        pcosh_piv = Mbs * math.log((math.exp(arg_piv) + math.exp(-arg_piv)) / 2.0)

    ten_pcosh_piv = math.pow(10., pcosh_piv)

    for idx in range(n):

        arg = _log10(x[idx] / be) / bs

        if arg < -6.0:

            pcosh = Mbs * (-arg - log2)

        elif arg > 4.0:

            pcosh = Mbs * (arg - log2)

        else:

            pcosh = Mbs * np.log(0.5 * ((np.exp(arg) + np.exp(-arg))))

        out[idx] = K * math.pow(x[idx]/piv, B) * \
            math.pow(10., pcosh)/ten_pcosh_piv

    return out


@nb.njit(fastmath=True, cache=_cache_functions)
def bb_eval(x, K, kT):

    return K * x * x / _expm1(x/kT)

@nb.njit(fastmath=True, cache=_cache_functions)
def mbb_eval(x, K, kT):

    arg = x/kT
    exp_arg = _exp(-arg)
    
    out = _pow(arg, 1.5) * exp_arg /_sqrt(1- exp_arg)
    return K * out / x

# @nb.njit(fastmath=True, cache=_cache_functions)
# def bbrad_eval(x, K, kT):

#     tinv = 1./kT
#     anorm = 1.0344E-3
#     anormh = 0.5*anorm

#     elow = 
    
#     xx = elow * tinv


    



# @nb.njit(fastmath=True, cache=_cache_functions)
# def bbrad_eval(x, K, kT):

#     n = x.shape[0]
#     out = np.empty(n)

#     for idx in range(n):

#         arg = x[idx]/kT
#         out[idx] = K * x[idx] * x[idx] / np.expm1(arg)

#     return out


# band calderone


@nb.njit(fastmath=True, cache=_cache_functions)
def ggrb_int_pl(a, b, Ec, Emin, Emax):

    pre = math.pow(a - b, a - b) * math.exp(b - a) / math.pow(Ec, b)

    if b != -2:
        b2 = 2+b

        return pre / (b2) * (math.pow(Emax, b2) - math.pow(Emin, b2))

    else:

        return pre * math.log(Emax/Emin)


# @nb.njit(fastmath=True, cache=_cache_functions)
# def ggrb_int_cpl(a, Ec, Emin, Emax):


@nb.njit(fastmath=True, cache=_cache_functions)
def non_diss_photoshere_generic(x, K, ec, piv, a, b):

    log_v = a * _log(x / piv) - _pow(x / ec, b)

    return K * _exp(log_v)


@nb.njit(fastmath=True, cache=_cache_functions)
def dbl_sbpl(x, K, a1, a2, b1, xp, xb, n1, n2, xpiv):

    xj = xp * _pow(-(a2 + 2)/ (b1 + 2), 1./((b1 - a2) * n2))

    arg1 = xj/xb
    arg2 = x/xb
    arg3 =  x/xj

    inner1 = _pow(arg2, -a1 * n1) + _pow(arg2, -a2 * n2)

    inner2 = _pow(arg1, -a1 * n1) + _pow(arg1, -a2 * n2)

    out = _pow(xb/xpiv, a1) * _pow( _pow(inner1, n2/n1) + _pow(arg3, -b1 * n2) * _pow(inner2, n2 / n1), -1/n2)

    return K * out
